/*
** File: dmabus.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent IOCTL handlers for ADMXRC3_*DMABus functions.
**
** (C) Copyright Alpha Data 2011-2012, 2015
*/

#include <df.h>
#include "device.h"
#include "dmabus.h"
#include "lock.h"
#include <ioctl_admxrc3.h>
#include <admxrc3/types.h>

#define DEBUGLEVEL_ERROR  (0) /* Level of debug messages relating to driver logic errors */
#define DEBUGLEVEL_CANCEL (1) /* Level of debug messages relating to cancellation / cleanup */
#define DEBUGLEVEL_NORMAL (6) /* Level of debug messages relating to normal operation */

typedef enum _State {
  StateInvalid       = 0,
  StateAcquireTicket = 1,
  StateDmaTransfer   = 2,
  StateCompleted     = 3
} State;

/* Forward declarations */
static void dmaBusCallback(CoreTicket* pTicket, void* pCallbackContext, CoreDmaStatus dmaStatus);
static void dmaBusCancelCallback(DfClientRequest* pRequest, DfIoStatus reason);
static void dmaBusTicketAcquiredCallback(void* pDevice, CoreTicket* pTicket, CoreAcquireStatus result);

/*
** -----------------------------------------------------------------------------
** Utility code
** -----------------------------------------------------------------------------
*/

#if 0
static const char*
getStateDescription(
  State state)
{
  switch (state) {
  case StateInvalid:
    return "Invalid";

  case StateAcquireTicket:
    return "AcquireTicket";

  case StateDmaTransfer:
    return "DmaTransfer";

  case StateCompleted:
    return "Completed";

  default:
    return "INVALID";
  }
}
#endif

static const char*
getDmaStatusDescription(
  CoreDmaStatus status)
{
  switch (status) {
  case CoreDmaSuccess:
    return "Success";

  case CoreDmaInvalidChannel:
    return "InvalidChannel";

  case CoreDmaInvalidRegion:
    return "InvalidRegion";

  case CoreDmaDirectionNotSupported:
    return "DirectionNotSupported";

  case CoreDmaHardwareError:
    return "HardwareError";

  case CoreDmaUnexpectedError:
    return "UnexpectedError";

  case CoreDmaInvalidTicket:
    return "InvalidTicket";

  case CoreDmaCancelled:
    return "Cancelled";

  case CoreDmaMappingFailure:
    return "MappingFailure";

  default:
    return "INVALID";
  }
}

static DfIoStatus
mapDmaStatus(
  CoreDmaStatus dmaStatus)
{
  switch (dmaStatus) {
  case CoreDmaSuccess:
    return DfIoStatusSuccess;

  case CoreDmaHardwareError:
    return DfIoStatusError(ADMXRC3_HARDWARE_ERROR);

  case CoreDmaInvalidChannel:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);

  case CoreDmaInvalidRegion:
    return DfIoStatusError(ADMXRC3_INVALID_REGION);

  case CoreDmaDirectionNotSupported:
    return DfIoStatusError(ADMXRC3_NOT_SUPPORTED);

  case CoreDmaCancelled:
    return DfIoStatusError(ADMXRC3_CANCELLED);

  case CoreDmaMappingFailure:
    return DfIoStatusError(ADMXRC3_NO_MEMORY);

  case CoreDmaUnexpectedError:
  case CoreDmaInvalidTicket:
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }
}

static DfIoStatus
mapAcquireStatus(
  CoreAcquireStatus acqStatus)
{
  switch (acqStatus) {
  case CoreAcquireSuccess:
    return DfIoStatusSuccess;

  case CoreAcquireMustWait:
    return DfIoStatusError(ADMXRC3_MUST_WAIT);

  case CoreAcquireCancelled:
    return DfIoStatusError(ADMXRC3_CANCELLED);

  case CoreAcquireRemoved:
  case CoreAcquireFailed:
    return DfIoStatusError(ADMXRC3_HARDWARE_ERROR);

  case CoreAcquireInvalidIndex:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);

  case CoreAcquireInvalid:
  case CoreAcquireUnexpectedError:
  case CoreAcquireWaiting:
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }
}

static void
completeRequest(
  DmaBusTicket* pTicket)
{
  DfClientRequest* pRequest = pTicket->pRequest;
  Admxrc3DeviceContext* pDevCtx = pTicket->pDevCtx;

  if (pTicket->bAcquired) {
    pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &pTicket->ticket);
  }
  dfRequestComplete(pRequest, pTicket->status);
  dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
}

DF_DECLARE_INLINE_FUNC(void, dereferenceTicket)(
  DmaBusTicket* pTicket)
{
  int usage;

  usage = dfAtomicDec(&pTicket->usage);
  dfAssert(usage >= 0);
  if (usage == 0) {
    completeRequest(pTicket);
  }
}

#if 0
DF_DECLARE_INLINE_FUNC(void, referenceTicket)(
  DmaBusTicket* pTicket)
{
  dfAtomicInc(&pTicket->usage);
}
#endif

/*
** -----------------------------------------------------------------------------
** Callbacks
** -----------------------------------------------------------------------------
*/

static void
dmaBusCancelCallback(
  DfClientRequest* pRequest,
  DfIoStatus reason)
{
  DmaBusTicket* pTicket = (DmaBusTicket*)dfRequestGetContext(pRequest);
  Admxrc3DeviceContext* pDevCtx = pTicket->pDevCtx;
  unsigned int dmaChannel = pTicket->dmaChannel;
  DfSpinLockFlags f;
  int state;

  f = dfSpinLockGet(&pTicket->lock);
  dfAssert(!pTicket->bCancel);
  pTicket->bCancel = TRUE;
  state = pTicket->state;
  switch (state) {
  case StateAcquireTicket:
    pTicket->status = DfIoStatusError(ADMXRC3_CANCELLED);
    dfSpinLockPut(&pTicket->lock, f);
    dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaBusCancelCallback: state=AcquireTicket cancelling acquisition\n"));
    pDevCtx->coreInterface.pAcquireCancel(pDevCtx->pCoreContext, &pTicket->ticket);
    break;

  case StateDmaTransfer:
    pTicket->status = DfIoStatusError(ADMXRC3_CANCELLED);
    dfSpinLockPut(&pTicket->lock, f);
    dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaBusCancelCallback: state=DmaTransfer cancelling DMA transfer\n"));
    if (!(pTicket->flags & ADMXRC3_DMA_QUEUETEST)) {
      pDevCtx->coreInterface.pDmaBusCancel(pDevCtx->pCoreContext, &pTicket->ticket, dmaChannel);
    }
    break;

  case StateCompleted:
    dfSpinLockPut(&pTicket->lock, f);
    dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaBusCancelCallback: state=Completed\n"));
    break;

  default:
    dfSpinLockPut(&pTicket->lock, f);
    dfDebugPrint(DEBUGLEVEL_ERROR, ("*** dmaBusCancelCallback: state=Invalid state %u\n", state));
    break;
  }
  dereferenceTicket(pTicket);
}

static void
dmaBusTicketAcquiredCallback(
  void* pDevice,
  CoreTicket* pCoreTicket,
  CoreAcquireStatus acqStatus)
{
  DmaBusTicket* pTicket = DF_CONTAINER_OF(pCoreTicket, DmaBusTicket, ticket);
  Admxrc3DeviceContext* pDevCtx = pTicket->pDevCtx;
  DfClientRequest* pRequest = pTicket->pRequest;
  unsigned int dmaChannel;
  CoreDmaStatus dmaStatus;
  DfSpinLockFlags f;

  dmaChannel = pTicket->dmaChannel;
  f = dfSpinLockGet(&pTicket->lock);
  dfAssert(pTicket->state == StateAcquireTicket);
  switch (acqStatus) {
  case CoreAcquireSuccess:
    if (pTicket->bCancel) {
      pTicket->bAcquired = TRUE;
      dfSpinLockPut(&pTicket->lock, f);
      dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaBusTicketAcquiredCallback: acqStatus=AcquireSuccess, cancel flag set, pRequest=%p\n", pRequest));
      dereferenceTicket(pTicket);
    } else {
      dmaStatus = pDevCtx->coreInterface.pDmaBusInitialize(pDevCtx->pCoreContext, &pTicket->ticket, dmaChannel, &pTicket->dmaParameters, dmaBusCallback, NULL);
      if (CoreDmaSuccess != dmaStatus) {
        pTicket->state = StateCompleted;
        dfSpinLockPut(&pTicket->lock, f);
        pTicket->status = mapDmaStatus(dmaStatus);
        /* Safe to release DMA engine because we set state to Completed and bCancel isn't set */
        pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &pTicket->ticket);
        if (dfRequestClearCallback(pRequest)) {
          dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaBusTicketAcquiredCallback: acqStatus=AcquireSuccess, start DMA failed, removed cancel function, pRequest=%p dmaStatus=%s\n",
            pRequest, getDmaStatusDescription(dmaStatus)));
          /* Cancel routine has been removed - it's safe to complete the request here */
          completeRequest(pTicket);
        } else {
          dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaBusTicketAcquiredCallback: acqStatus=AcquireSuccess, start DMA failed, failed to remove cancel function, pRequest=%p dmaStatus=%s\n"));
          dereferenceTicket(pTicket);
        }
      } else {
        pTicket->bAcquired = TRUE;
        pTicket->state = StateDmaTransfer;
        dfSpinLockPut(&pTicket->lock, f);
        dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaBusTicketAcquiredCallback: acqStatus=AcquireSuccess, starting DMA, pRequest=%p\n", pRequest));
        if (pTicket->flags & ADMXRC3_DMA_QUEUETEST) {
          dmaBusCallback(&pTicket->ticket, NULL, CoreDmaSuccess);
        } else {
          pDevCtx->coreInterface.pDmaBusStart(pDevCtx->pCoreContext, &pTicket->ticket, dmaChannel);
        }
      }
    }
    break;

  case CoreAcquireMustWait:
    if (pTicket->bCancel) {
      dfSpinLockPut(&pTicket->lock, f);
      dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaBusTicketAcquiredCallback: acqStatus=AcquireMustWait, cancel flag set, pRequest=%p\n", pRequest));
      dereferenceTicket(pTicket);
    } else {
      pTicket->state = StateCompleted;
      dfSpinLockPut(&pTicket->lock, f);
      pTicket->status = mapAcquireStatus(acqStatus);
      if (dfRequestClearCallback(pRequest)) {
        dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaBusTicketAcquiredCallback: acqStatus=AcquireMustWait, removed cancel function, pRequest=%p\n", pRequest));
        completeRequest(pTicket);
      } else {
        dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaBusTicketAcquiredCallback: acqStatus=AcquireMustWait, failed to remove cancel function, pRequest=%p\n", pRequest));
        dereferenceTicket(pTicket);
      }
    }
    break;

  case CoreAcquireCancelled:
    dfSpinLockPut(&pTicket->lock, f);
    dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaBusTicketAcquiredCallback: acqStatus=AcquireCancelled, pRequest=%p\n", pRequest));
    dereferenceTicket(pTicket);
    break;

  default:
    if (pTicket->bCancel) {
      dfSpinLockPut(&pTicket->lock, f);
      dfDebugPrint(DEBUGLEVEL_ERROR, ("*** dmaBusTicketAcquiredCallback: acqStatus=Acquire??, cancel flag set, pRequest=%p\n", pRequest));
      dereferenceTicket(pTicket);
    } else {
      pTicket->state = StateCompleted;
      dfSpinLockPut(&pTicket->lock, f);
      pTicket->status = mapAcquireStatus(acqStatus);
      if (dfRequestClearCallback(pRequest)) {
        dfDebugPrint(DEBUGLEVEL_ERROR, ("dmaBusTicketAcquiredCallback: acqStatus=Acquire??, removed cancel function, pRequest=%p\n", pRequest));
        completeRequest(pTicket);
      } else {
        dfDebugPrint(DEBUGLEVEL_ERROR, ("dmaBusTicketAcquiredCallback: acqStatus=Acquire??, failed to remove cancel function, pRequest=%p\n", pRequest));
        dereferenceTicket(pTicket);
      }
    }
    break;
  }
}

static void
dmaBusCallback(
  CoreTicket* pCoreTicket,
  void* pCallbackContext, /* ignored */
  CoreDmaStatus dmaStatus)
{
  DmaBusTicket* pTicket = DF_CONTAINER_OF(pCoreTicket, DmaBusTicket, ticket);
	Admxrc3DeviceContext* pDevCtx = pTicket->pDevCtx;
  DfClientRequest* pRequest = pTicket->pRequest;
  DfSpinLockFlags f;

  f = dfSpinLockGet(&pTicket->lock);
  dfAssert(pTicket->state == StateDmaTransfer);
  if (pTicket->bCancel) {
    dfSpinLockPut(&pTicket->lock, f);
    dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaBusCallback: Cancel flag set, pRequest=%p\n", pRequest));
    dereferenceTicket(pTicket);
  } else {
    pTicket->state = StateCompleted;
    pTicket->bAcquired = FALSE;
    dfSpinLockPut(&pTicket->lock, f);
    pTicket->status = mapDmaStatus(dmaStatus);
    /* Safe to release DMA engine because we set state to Completed */
    pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &pTicket->ticket);
    if (dfRequestClearCallback(pRequest)) {
      dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaBusCallback: Removed cancel function, pRequest=%p dmaStatus=%s\n",
        pRequest, getDmaStatusDescription(dmaStatus)));
      /* Cancel routine has been removed - it's safe to complete the request here */
      completeRequest(pTicket);
    } else {
      dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaBusCallback: Failed to remove cancel function, pRequest=%p\n", pRequest));
      dereferenceTicket(pTicket);
    }
  }
}

/*
** -----------------------------------------------------------------------------
** IOCTL handlers
** -----------------------------------------------------------------------------
*/

#if defined(ADMXRC3_CHECK_FLAGS) && ADMXRC3_CHECK_FLAGS
static const uint32_t legalFlagsMask = ADMXRC3_DMA_FIXEDLOCAL | ADMXRC3_DMA_DONOTQUEUE | ADMXRC3_DMA_QUEUETEST | ADMXRC3_DMA_BYPASSHW;
#endif

DfIoStatus
ioctlReadDmaBus(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_READDMABUS* pIoctl = (IOCTLS_ADMXRC3_READDMABUS*)pBuffer;
  DmaBusTicket* pTicket = NULL;
  CoreAcquireStatus acqStatus;
  unsigned int dmaChannel;
  size_t length;
  uint32_t acquireFlags, flags;
  uint64_t localAddress, busAddress;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  dmaChannel = pIoctl->in.channel;
  if (dmaChannel >= pDevCtx->coreInterface.pGetNumDmaChannel(pDevCtx->pCoreContext)) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  length = pIoctl->in.length;
  if (0 == length) {
    dfRequestPending(pRequest);
    dfRequestComplete(pRequest, DfIoStatusSuccess);
    return DfIoStatusDeferred;
  }
  flags = pIoctl->in.flags;
  localAddress = pIoctl->in.local;
  busAddress = pIoctl->in.bus;
#if defined(ADMXRC3_CHECK_FLAGS) && ADMXRC3_CHECK_FLAGS
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif
  if (!(flags & ADMXRC3_DMA_FIXEDLOCAL)) {
    if (localAddress + length < localAddress || localAddress + length > pDevCtx->localAddressLimit) {
      return DfIoStatusError(ADMXRC3_INVALID_LOCAL_REGION);
    }
  } else {
    uint8_t tmp = (uint8_t)(pDevCtx->localWidth - (localAddress & (pDevCtx->localWidth - 1U)));

    if (localAddress + length < localAddress || localAddress + tmp > pDevCtx->localAddressLimit) {
      return DfIoStatusError(ADMXRC3_INVALID_LOCAL_REGION);
    }
  }
  acquireFlags = (flags & ADMXRC3_DMA_DONOTQUEUE) ? CORE_ACQUIRE_NOWAIT : 0;
  pTicket = dfPoolAlloc(&pDevCtx->dma.ticketPool, DmaBusTicket);
  if (NULL == pTicket) {
    return DfIoStatusError(ADMXRC3_NO_MEMORY);
  }

  dfSpinLockInit(&pTicket->lock);
  dfAtomicSet(&pTicket->usage, 2);
  pTicket->pRequest = pRequest;
  pTicket->pDevCtx = pDevCtx;
  pTicket->pClCtx = pClCtx;
  pTicket->dmaChannel = dmaChannel;
  pTicket->flags = flags;
  pTicket->dmaParameters.bWriteToLocal = FALSE;
  pTicket->dmaParameters.flags = flags & (CoreDmaFlagFixedLocal | CoreDmaFlagBypassHardware);
  pTicket->dmaParameters.region.offset = 0;
  pTicket->dmaParameters.region.length = length;
  pTicket->dmaParameters.region.localAddress = localAddress;
  pTicket->dmaParameters.region.busAddress = busAddress;
  pTicket->bAcquired = FALSE;
  pTicket->bCancel = FALSE;
  pTicket->status = DfIoStatusSuccess;
  pTicket->state = StateAcquireTicket;
  pTicket->ticket.count = 1;
  pTicket->ticket.resources[0].code = CORE_RESOURCE_DMAENGINE;
  pTicket->ticket.resources[0].index = (uint8_t)dmaChannel;
  dfRequestSetContext(pRequest, pTicket);

  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &pTicket->ticket);
  if (acqStatus != CoreAcquireSuccess) {
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    return mapAcquireStatus(acqStatus);
  }

  dfRequestPending(pRequest);

  if (!dfRequestSetCallback(pRequest, dmaBusCancelCallback)) {
    dfRequestComplete(pRequest, DfIoStatusError(ADMXRC3_CANCELLED));
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    return DfIoStatusDeferred;
  }

  pDevCtx->coreInterface.pAcquireAsync(pDevCtx->pCoreContext, &pTicket->ticket, acquireFlags, dmaBusTicketAcquiredCallback);
  return DfIoStatusDeferred;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlReadDmaBusThunk(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_READDMABUS* pIoctl32 = (IOCTLS32_ADMXRC3_READDMABUS*)pBuffer;
  IOCTLS_ADMXRC3_READDMABUS ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.channel = pIoctl32->in.channel;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.bus = pIoctl32->in.bus;
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.local = pIoctl32->in.local;
  return ioctlReadDmaBus(pDevCtx, pClCtx, pRequest, &ioctl, sizeof(ioctl.in));
}
#endif

DfIoStatus
ioctlWriteDmaBus(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_READDMABUS* pIoctl = (IOCTLS_ADMXRC3_READDMABUS*)pBuffer;
  DmaBusTicket* pTicket = NULL;
  CoreAcquireStatus acqStatus;
  unsigned int dmaChannel;
  size_t length;
  uint32_t acquireFlags, flags;
  uint64_t localAddress, busAddress;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  dmaChannel = pIoctl->in.channel;
  if (dmaChannel >= pDevCtx->coreInterface.pGetNumDmaChannel(pDevCtx->pCoreContext)) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  length = pIoctl->in.length;
  if (0 == length) {
    dfRequestPending(pRequest);
    dfRequestComplete(pRequest, DfIoStatusSuccess);
    return DfIoStatusDeferred;
  }
  flags = pIoctl->in.flags;
  localAddress = pIoctl->in.local;
  busAddress = pIoctl->in.bus;
#if defined(ADMXRC3_CHECK_FLAGS) && ADMXRC3_CHECK_FLAGS
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif
  if (!(flags & ADMXRC3_DMA_FIXEDLOCAL)) {
    if (localAddress + length < localAddress || localAddress + length > pDevCtx->localAddressLimit) {
      return DfIoStatusError(ADMXRC3_INVALID_LOCAL_REGION);
    }
  } else {
    uint8_t tmp = (uint8_t)(pDevCtx->localWidth - (localAddress & (pDevCtx->localWidth - 1U)));

    if (localAddress + length < localAddress || localAddress + tmp > pDevCtx->localAddressLimit) {
      return DfIoStatusError(ADMXRC3_INVALID_LOCAL_REGION);
    }
  }
  acquireFlags = (flags & ADMXRC3_DMA_DONOTQUEUE) ? CORE_ACQUIRE_NOWAIT : 0;
  pTicket = dfPoolAlloc(&pDevCtx->dma.ticketPool, DmaBusTicket);
  if (NULL == pTicket) {
    return DfIoStatusError(ADMXRC3_NO_MEMORY);
  }

  dfSpinLockInit(&pTicket->lock);
  dfAtomicSet(&pTicket->usage, 2);
  pTicket->pRequest = pRequest;
  pTicket->pDevCtx = pDevCtx;
  pTicket->pClCtx = pClCtx;
  pTicket->dmaChannel = dmaChannel;
  pTicket->flags = flags;
  pTicket->dmaParameters.bWriteToLocal = TRUE;
  pTicket->dmaParameters.flags = flags & (CoreDmaFlagFixedLocal | CoreDmaFlagBypassHardware);
  pTicket->dmaParameters.region.offset = 0;
  pTicket->dmaParameters.region.length = length;
  pTicket->dmaParameters.region.localAddress = localAddress;
  pTicket->dmaParameters.region.busAddress = busAddress;
  pTicket->bAcquired = FALSE;
  pTicket->bCancel = FALSE;
  pTicket->status = DfIoStatusSuccess;
  pTicket->state = StateAcquireTicket;
  pTicket->ticket.count = 1;
  pTicket->ticket.resources[0].code = CORE_RESOURCE_DMAENGINE;
  pTicket->ticket.resources[0].index = (uint8_t)dmaChannel;
  dfRequestSetContext(pRequest, pTicket);

  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &pTicket->ticket);
  if (acqStatus != CoreAcquireSuccess) {
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    return mapAcquireStatus(acqStatus);
  }

  dfRequestPending(pRequest);

  if (!dfRequestSetCallback(pRequest, dmaBusCancelCallback)) {
    dfRequestComplete(pRequest, DfIoStatusError(ADMXRC3_CANCELLED));
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    return DfIoStatusDeferred;
  }

  pDevCtx->coreInterface.pAcquireAsync(pDevCtx->pCoreContext, &pTicket->ticket, acquireFlags, dmaBusTicketAcquiredCallback);
  return DfIoStatusDeferred;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlWriteDmaBusThunk(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_WRITEDMABUS* pIoctl32 = (IOCTLS32_ADMXRC3_WRITEDMABUS*)pBuffer;
  IOCTLS_ADMXRC3_WRITEDMABUS ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.channel = pIoctl32->in.channel;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.bus = pIoctl32->in.bus;
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.local = pIoctl32->in.local;
  return ioctlWriteDmaBus(pDevCtx, pClCtx, pRequest, &ioctl, sizeof(ioctl.in));
}
#endif
