/*
** File: model_adpexrc6t.c
** Project: ADMXRC3 module driver
** Purpose: ADPE-XRC-6T specific code, including the function to bootstrap
**          the device context.
**
** (C) Copyright Alpha Data 2011
*/

#include <df.h>

#include "device.h"
#include "info.h"
#include "model_boot.h"
#include <core/adpexrc6t.h>

static ADMXRC3_STATUS
getBankInfo(
	Admxrc3DeviceContext* pDevCtx,
  unsigned int bankIndex,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETBANKINFO* pIoctl)
{
  VpdAdpexrc6tRev0* pVpd = (VpdAdpexrc6tRev0*)pVpdBuffer;
  uint8_t order;

  if (bankIndex >= 2) {
    return ADMXRC3_INVALID_INDEX;
  }
  order = pVpd->sdram[bankIndex].order;
  pIoctl->out.typeMask = ADMXRC3_BANK_SDRAM_DDR3;
  pIoctl->out.physicalDataWidth = 32;
  pIoctl->out.physicalEccWidth = 0;
  pIoctl->out.maximumFrequency = (uint64_t)pVpd->sdram[bankIndex].speed * 100000;
  pIoctl->out.minimumFrequency = 303030303U;
  pIoctl->out.connectivityMask = 0x1U;
  pIoctl->out.physicalSize = (uint64_t)0x1U << order;
  pIoctl->out.bPresent = (uint8_t)(order != 0);

  return ADMXRC3_SUCCESS;
}

static void
getCardInfo(
	Admxrc3DeviceContext* pDevCtx,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETCARDINFO* pIoctl)
{
  VpdAdpexrc6tRev0* pVpd = (VpdAdpexrc6tRev0*)pVpdBuffer;
  unsigned int i;

  pIoctl->out.serialNumber = pVpd->serial;
  pIoctl->out.numMemoryBank = 2;
  pIoctl->out.memoryBankPresent = 0;
  for (i = 0; i < 2; i++) {
    if (0 != pVpd->sdram[i].order) {
      /* Bank is fitted */
      pIoctl->out.memoryBankPresent |= (0x1U << i);
    }
  }
}

static ADMXRC3_STATUS
getFpgaInfo(
	Admxrc3DeviceContext* pDevCtx,
  unsigned int targetIndex,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETFPGAINFO* pIoctl)
{
  VpdAdpexrc6tRev0* pVpd = (VpdAdpexrc6tRev0*)pVpdBuffer;
  ADMXRC3_PACKAGE_TYPE package;
  ADMXRC3_FPGA_TYPE fpgaType;
  char speed[8];
  char stepping[8];

  switch (targetIndex) {
  case 0:
    package = ADMXRC3_PACKAGE_MAKEFF(1156);
    fpgaType = (ADMXRC3_FPGA_TYPE)pVpd->fpga.device;
    break;

  default:
    return ADMXRC3_INVALID_INDEX;
  }

  dfAssert(sizeof(char) == sizeof(uint8_t));

  dfStrCpy(speed, DF_ARRAY_LENGTH(speed), (const char*)pVpd->fpga.speedGrade);
  switch (pVpd->tempGrade) {
  case ADPEXRC6T_VPD_TEMPGRADE_COMMERCIAL:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "C");
    break;

  case ADPEXRC6T_VPD_TEMPGRADE_INDUSTRIAL:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "I");
    break;

  case ADPEXRC6T_VPD_TEMPGRADE_EXTENDED:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "E");
    break;

  default:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "?");
    break;
  }

  dfStrCpy(stepping, DF_ARRAY_LENGTH(stepping), (const char*)pVpd->fpga.stepping);

  pIoctl->out.flags = ADMXRC3_FPGA_SPEEDVALID | ADMXRC3_FPGA_STEPPINGVALID;
  setupIoctlsGetFpgaInfo(pDevCtx, pIoctl, fpgaType, package, speed, stepping, TRUE);

  return ADMXRC3_SUCCESS;
}

void
admxrc3BootstrapAdpexrc6t(
	Admxrc3DeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("admxrc3BootstrapAdpexrc6t: entered\n"));

  pDevCtx->methods.pGetBankInfo = getBankInfo;
  pDevCtx->methods.pGetCardInfo = getCardInfo;
  pDevCtx->methods.pGetFpgaInfo = getFpgaInfo;
  pDevCtx->bReverseSelectMapData = FALSE;
  pDevCtx->localAddressLimit = (uint64_t)(0x1ULL << 39U);
  pDevCtx->localWidth = 16U;
}
